-- by mor2000

-- needs getMinMaxKeys(), round(), times10[]

local TYPE_NIL = 0
local TYPE_TRUE = 1
local TYPE_FALSE = 2
local TYPE_INT6 = 3
local TYPE_INT12 = 4
local TYPE_INT18 = 5
local TYPE_INT32 = 6
local TYPE_INT64 = 7
local TYPE_FLOAT6 = -1
local TYPE_FLOAT12 = -2
local TYPE_FLOAT18 = -3
local TYPE_FLOAT32 = -4
local TYPE_FLOAT64 = -5
local TYPE_STRING = -6
local TYPE_LIST = -7

local STREAMTYPEBITS = 4 -- max type <= 2^X +1

function _m.streamWriteAuto(streamId, ...)
	local args = {...}
	local _,argslen=getMinMaxKeys(args)
	if (argslen or 0)>1 then
		streamWriteIntN(streamId, TYPE_LIST, STREAMTYPEBITS)
		streamWriteAuto(streamId, argslen)
		for i=1,argslen do
			streamWriteAuto(streamId, args[i])
		end
		return
	end
	arg = args[1]
	local argtype = type(arg)
	if arg==true then
		streamWriteIntN(streamId, TYPE_TRUE, STREAMTYPEBITS)
	elseif arg==false then
		streamWriteIntN(streamId, TYPE_FALSE, STREAMTYPEBITS)
	elseif argtype=="number" then
		local arground = round(arg,8)
		local argabs = math.abs(arground)
		if arground%1==0 then
			-- int
			if argabs<32 then
				streamWriteIntN(streamId, TYPE_INT6, STREAMTYPEBITS)
				streamWriteIntN(streamId, arground, 6)
			elseif argabs<2048 then
				streamWriteIntN(streamId, TYPE_INT12, STREAMTYPEBITS)
				streamWriteIntN(streamId, arground, 12)
			elseif argabs<131072 then
				streamWriteIntN(streamId, TYPE_INT18, STREAMTYPEBITS)
				streamWriteIntN(streamId, arground, 18)
			elseif argabs<2147483648 then
				streamWriteIntN(streamId, TYPE_INT32, STREAMTYPEBITS)
				streamWriteInt32(streamId, arground)
			else
				local first = math.floor(arground/(2^31)) -- 32-1
				local second = arground-first*2^31
				streamWriteIntN(streamId, TYPE_INT64, STREAMTYPEBITS)
				streamWriteInt32(streamId, first)
				streamWriteInt32(streamId, second)
			end
		else
			-- float
			local intval = math.floor(arground)
			local intvalabs = math.abs(intval)
			if intvalabs<32 then
				streamWriteIntN(streamId, TYPE_FLOAT6, STREAMTYPEBITS)
				streamWriteIntN(streamId, intval, 6)
			elseif intvalabs<2048 then
				streamWriteIntN(streamId, TYPE_FLOAT12, STREAMTYPEBITS)
				streamWriteIntN(streamId, intval, 12)
			elseif intvalabs<131072 then
				streamWriteIntN(streamId, TYPE_FLOAT18, STREAMTYPEBITS)
				streamWriteIntN(streamId, intval, 18)
			elseif intvalabs<2147483648 then
				streamWriteIntN(streamId, TYPE_FLOAT32, STREAMTYPEBITS)
				streamWriteInt32(streamId, intval)
			else
				local first = math.floor(intval/(2^31)) -- 32-1
				local second = intval-first*2^31
				streamWriteIntN(streamId, TYPE_FLOAT64, STREAMTYPEBITS)
				streamWriteInt32(streamId, first)
				streamWriteInt32(streamId, second)
			end
			local decimalplaces = math.floor(arground*times10[8]-intval*times10[8])
			streamWriteInt32(streamId, decimalplaces)
		end
	elseif argtype=="string" then
		-- string
		streamWriteIntN(streamId, TYPE_STRING, STREAMTYPEBITS)
		streamWriteString(streamId, arg)
	else
		-- nil or else
		streamWriteIntN(streamId, TYPE_NIL, STREAMTYPEBITS)
	end
end

function _m.streamReadAuto(streamId)
	local rettype = streamReadIntN(streamId, STREAMTYPEBITS)
	local ret = nil
	if rettype==TYPE_LIST then
		local rets={}
		local retslen = streamReadAuto(streamId)
		for i=1,retslen do
			rets[i] = streamReadAuto(streamId)
		end
		return unpack(rets)
	elseif rettype==TYPE_TRUE then
		ret = true
	elseif rettype==TYPE_FALSE then
		ret = false
	elseif rettype==TYPE_INT6 then
		ret = streamReadIntN(streamId, 6)
	elseif rettype==TYPE_INT12 then
		ret = streamReadIntN(streamId, 12)
	elseif rettype==TYPE_INT18 then
		ret = streamReadIntN(streamId, 18)
	elseif rettype==TYPE_INT32 then
		ret = streamReadInt32(streamId)
	elseif rettype==TYPE_INT64 then
		local first = streamReadInt32(streamId)
		local second = streamReadInt32(streamId)
		ret = first*2^31+second
	elseif rettype==TYPE_FLOAT6 then
		local intval = streamReadIntN(streamId,6)
		local decimalplaces = streamReadInt32(streamId)
		ret = intval + decimalplaces/times10[8]
	elseif rettype==TYPE_FLOAT12 then
		local intval = streamReadIntN(streamId,12)
		local decimalplaces = streamReadInt32(streamId)
		ret = intval + decimalplaces/times10[8]
	elseif rettype==TYPE_FLOAT18 then
		local intval = streamReadIntN(streamId,18)
		local decimalplaces = streamReadInt32(streamId)
		ret = intval + decimalplaces/times10[8]
	elseif rettype==TYPE_FLOAT32 then
		local intval = streamReadInt32(streamId)
		local decimalplaces = streamReadInt32(streamId)
		ret = intval + decimalplaces/times10[8]
	elseif rettype==TYPE_FLOAT64 then
		local first = streamReadInt32(streamId)
		local second = streamReadInt32(streamId)
		local intval = first*2^31+second
		local decimalplaces = streamReadInt32(streamId)
		ret = intval + decimalplaces/times10[8]
	elseif rettype==TYPE_STRING then
		ret = streamReadString(streamId)
	else
		-- nil
	end
	return ret
end
